import os

from pytest import Config

from pytest_testrail_integrator.constants import TestrailMsgStyle


class TrConfig:
    """Object storing entire configs set for Testrail-Integrator plugin."""

    def __init__(self, config: Config):
        self._config = config
        self.start_client = config.getoption('--tr-reporting')
        self.deselect_tests = config.getoption('--tr-deselect-tests', False)
        self.run_id = os.getenv('TR_RUN_ID') or config.getini('tr_run_id')
        self.api_url = os.getenv('TR_API_URL') or config.getini('tr_api_url')
        self.user_email = os.getenv('TR_USER_EMAIL') or config.getini('tr_user_email')
        self.user_password = os.getenv('TR_USER_PASSWORD') or config.getini('tr_user_password')
        self.project_id = os.getenv('TR_PROJECT_ID') or config.getini('tr_project_id')
        self.suite_id = os.getenv('TR_SUITE_ID') or config.getini('tr_suite_id')
        self.tb_style = TestrailMsgStyle(config.getoption('--tr-tb') or config.getini('tr_tb') or 'short')

        required_keys = (self.api_url, self.user_email, self.user_password)
        assert all(required_keys)



import pickle

from pytest_testrail_integrator.client import TrClient, WORKER_RESULTS_KEY


class XdistTrClintAdapter:
    def __init__(self, tr_client: TrClient):
        self._tr_client = tr_client

    def pytest_testnodedown(self, node, error):
        """Extend master node results with worker's node results."""
        node_reports = pickle.loads(node.workeroutput[WORKER_RESULTS_KEY])
        self._tr_client.add_results(node_reports)



from typing import List, Iterable

from testrail_api import TestRailAPI

from pytest_testrail_integrator.config import TrConfig


class TrService:
    _tests: list

    def __init__(self, config, tr_api: TestRailAPI):
        self._config: TrConfig = config.tr_config
        self._api = tr_api

    def is_test_run_available(self) -> bool:
        return bool(self._config.run_id)

    @property
    def tests(self):
        """Gets Tests from run if RunId is defined.

        Returns:
            List of test dicts.
        Raises:
            ValueError: When runId in configs is not defined.

        """
        if not hasattr(self, '_tests'):
            if not self._config.run_id:
                raise ValueError('RunId is not available.')
            # Todo: Add support of receiving more than 250 tests
            self._tests = self._api.tests.get_tests(int(self._config.run_id))['tests']
        return self._tests

    def get_cases(self) -> List[int]:
        """Gets list of case ids within entire test run if defined.

        Returns:
            List of integers.

        """
        return [x['case_id'] for x in self.tests]

    def upload_results(self, results):
        if results:
            return self._api.results.add_results_for_cases(int(self._config.run_id), results)

    def create_test_run(self, run_name: str,  case_ids: Iterable[int]) -> None:
        """Creates new test with test cases matching with passed case_ids param.

        Args:
            run_name: New test run name str.
            case_ids: List of int, testrail case ids.

        Returns:
            None

        """
        assert self._config.suite_id and self._config.suite_id.isdigit()

        tr_cases = self.__filter_pytest_cases(case_ids)

        r = self._api.runs.add_run(
            self._config.project_id,
            name=run_name,
            include_all=False,
            case_ids=tr_cases,
            suite_id=self._config.suite_id
        )
        self._config.run_id = r['id']

    def __filter_pytest_cases(self, cases: Iterable[int]) -> List[int]:
        """
        Filter cases and return the only cases which are present in testrail suite.

        Args:
            cases: List of integers. Testcase ids.

        Returns:
            List[int] containing the only ids present in testrail.

        """
        # Todo: Add support of getting more than 250 cases.
        suite_cases = self._api.cases.get_cases(self._config.project_id, suite_id=self._config.suite_id)['cases']
        suite_cases_ids = [x['id'] for x in suite_cases]

        return list(filter(lambda x: x in suite_cases_ids, cases))



import pytest
from pytest import PytestPluginManager, Parser
from testrail_api import TestRailAPI

from pytest_testrail_integrator.client import TrClient
from pytest_testrail_integrator.config import TrConfig
from pytest_testrail_integrator.service import TrService


def pytest_addoption(parser: Parser):
    group = parser.getgroup('testrail')
    group.addoption(
        '--tr-reporting',
        action='store_true',
        help='Setup TestRail report client for upload test results.'
    )
    group.addoption(
        '--tr-deselect-tests',
        action='store_true',
        help='Deselect tests if its\' case is not present in run.'
    )
    group.addoption(
        '--tr-tb',
        action='store',
        help='Sets traceback level in testrail message reports.'
    )
    parser.addini('tr_url', default='', help='Testrail Api url.')
    parser.addini('tr_user_email', default='', help='Testrail user\'s email.')
    parser.addini('tr_user_password', default='', help='Testrail user\'s password.')
    parser.addini('tr_project_id', default='', help='Testrail project id for new testrun creation.')
    parser.addini('tr_suite_id', default='', help='Testrail suite id to create test run from.')
    parser.addini('tr_run_id', default='', help='Testrail run id to update tests in.')
    parser.addini('tr_tb', default='', help='Sets traceback level in testrail message reports.')


@pytest.hookimpl
def pytest_addhooks(pluginmanager: PytestPluginManager):
    from pytest_testrail_integrator import new_hooks
    pluginmanager.add_hookspecs(new_hooks)


def pytest_configure(config):
    if config.getoption('--tr-reporting'):
        tr_config = TrConfig(config)
        tr_api = TestRailAPI(tr_config.api_url, tr_config.user_email, tr_config.user_password)
        config.tr_config = tr_config
        config.tr_service = TrService(config, tr_api)
        plugin = TrClient(config)
        config.pluginmanager.register(plugin)
        if config.pluginmanager.hasplugin("xdist"):

            from pytest_testrail_integrator.xdist.xdist_tr_adapter import XdistTrClintAdapter
            config.pluginmanager.register(XdistTrClintAdapter(plugin))

        config.testrail_reporter = plugin



import pickle
from datetime import datetime
from typing import List, Union, Iterable

import pydash
import pytest
from pytest import ExitCode, Config, Session, Mark, Item, TestReport, CallInfo

from .config import TrConfig
from .constants import TR_MARKER_NAME, TR_PASSED_TESTS_FLUSH_SIZE, TestrailStatus, \
    PytestStatus, TESTRAIL_STATUS_PRIORITY, PYTEST_TO_TESTRAIL_STATUS, TestrailMsgStyle
from .dto import ReportDTO
from .service import TrService

WORKER_RESULTS_KEY = 'tr_results'


def tr_case(case_id: Union[str, int]):
    return pytest.mark.case(case_id)


def is_master(config: Config) -> bool:
    """
    True if the code running the given pytest.config object is
    running in a xdist master node or not running xdist at all.
    """
    return not hasattr(config, 'workerinput')


class TrClient:
    def __init__(self, config):
        self._service: TrService = config.tr_service
        self._config = config
        self._tr_config: TrConfig = config.tr_config
        self._results: List[ReportDTO] = []
        self._passed_tests_count = 0

    def add_results(self, results: Iterable[ReportDTO]) -> None:
        self._results.extend(results)

    def pytest_collection_modifyitems(self, session: Session, config: Config, items: List[Item]):
        """Select tests containing in testrail test run only."""
        cond = (self._service.is_test_run_available(), self._tr_config.deselect_tests)
        if not all(cond):
            return

        tr_run_cases = self._service.get_cases()
        deselected, selected = [], []

        for item in items:
            marker = item.get_closest_marker(TR_MARKER_NAME)
            if not marker:
                deselected.append(item)
                continue

            case_id = self._get_marker_case_id(marker)
            if case_id not in tr_run_cases:
                deselected.append(item)
            else:
                selected.append(item)
        items[:] = selected

        config.hook.pytest_deselected(items=deselected)

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_makereport(self, item: Item, call: CallInfo):
        marker = item.get_closest_marker(TR_MARKER_NAME)
        if not marker:
            return

        case_id = self._get_marker_case_id(marker)

        result: TestReport = (yield).get_result()

        if call.when == 'call' or result.outcome == 'skipped':

            report = ReportDTO(
                item.name,
                item.nodeid,
                PytestStatus(result.outcome),
                result.duration,
                result.longrepr,
                case_id,
                self.__is_parametrized_test(item.own_markers)
            )

            self._results.append(report)

            if report.status == PytestStatus.PASSED:
                self._passed_tests_count += 1

            # Try to flush passed results per node
            self._try_flush_reports()

    @staticmethod
    def _get_marker_case_id(marker):
        case = marker.args[0]
        if isinstance(case, str):
            case = case.lower().replace('c', '')
        case_id = int(case)
        return case_id

    def _try_flush_reports(self):
        if self._service.is_test_run_available() and \
                self._passed_tests_count > TR_PASSED_TESTS_FLUSH_SIZE:
            self._send_passed_reports(self._results)
            self._passed_tests_count = 0

    @pytest.hookimpl(hookwrapper=True, trylast=True)
    def pytest_sessionfinish(self, session: Session, exitstatus: Union[int, ExitCode]):
        yield
        if not is_master(session.config):
            # Serialize results on worker node and place in workeroutput to pass results to master node
            workeroutput = getattr(session.config, "workeroutput")
            workeroutput[WORKER_RESULTS_KEY] = pickle.dumps(self._results)
        else:
            if not self._service.is_test_run_available():
                # Todo: Probably it would be better to create new test run after all tests are collected.
                config = session.config
                run_name = config.hook.pytest_tr_generate_run_name(config=config)
                actual_cases = {x.case_id for x in self._results}
                self._service.create_test_run(run_name, actual_cases)

            self._service.upload_results(self._prepare_report(self._results, self._service.get_cases()))

    @pytest.hookimpl(trylast=True)
    def pytest_tr_generate_run_name(self, config):
        """Default tr run name generation."""
        current_date = datetime.now().utcnow().strftime("%d-%h-%y %H:%MUTC")
        return 'Automated test run ' + current_date

    def _prepare_report(self, results: List[ReportDTO], testrun_cases: List[int]):
        # Remove redundant result reports
        filtered = filter(lambda x: x.case_id in testrun_cases, results)
        out = []
        for report_item in filtered:
            out.append(self.__prepare_item_report(report_item))

        # Sort results by testrail status to get failed results on top.
        return sorted(out, key=lambda x: TESTRAIL_STATUS_PRIORITY[TestrailStatus(x['status_id'])])

    def __prepare_item_report(self, report: ReportDTO) -> dict:
        """Prepares API report item."""

        return {
            "case_id": report.case_id,
            "status_id": PYTEST_TO_TESTRAIL_STATUS[report.status].value,
            "comment": self.__prepare_comment(report),
            "elapsed": f"{round(report.duration, 2) or 0.01}s"  # No need to pass more precise elapse time than 0.01s
        }

    def __prepare_comment(self, report: ReportDTO):
        msg = report.name
        if report.is_parametrized:
            # crop by [ and ] to select parametrize ID
            msg = f"Test ID: {msg[msg.find('[') + 1: msg.find(']')]}"

        if report.status == PytestStatus.FAILED:
            tb = f'{msg}\n{pydash.get(report, "longrepr.reprcrash.message", "")}'
            if self._tr_config.tb_style == TestrailMsgStyle.LONG:
                return tb
            elif self._tr_config.tb_style == TestrailMsgStyle.SHORT:
                return self._shorten_error(tb)
        if report.status == PytestStatus.SKIPPED:
            return msg + f'\n{pydash.get(report, "longrepr[2]", "")}'
        return msg

    @staticmethod
    def _shorten_error(err: str) -> str:
        """Cuts long traceback message."""
        if not err:
            return err

        cut_until = err.find('\n + ')  # May be changed to more accurate substr
        return err[:cut_until]

    @staticmethod
    def __is_parametrized_test(markers: List[Mark]):
        """
        Checks if the list contains parametrized marker.

        Args:
            markers: List of test's markers.

        Returns:
            boolean

        """

        return any([x.name == 'parametrize' for x in markers])

    def _send_passed_reports(self, results: List[ReportDTO]) -> None:
        """Prepare and send all passed tests only then delete them from the reports list.

        Args:
            results: All results within node.

        Returns:
            None

        """
        passed_results = list(filter(lambda x: x.status == PytestStatus.PASSED, results))
        results[:] = list(filter(lambda x: x.status != PytestStatus.PASSED, results))

        self._service.upload_results(self._prepare_report(passed_results, self._service.get_cases()))
